/**
   @author Shin'ichiro NAKAOKA
*/

#include "ExtensionManager.h"
#include "ItemManager.h"
#include "MenuManager.h"
#include "OptionManager.h"
#include "ProjectManager.h"
#include "Item.h"
#include "View.h"
#include "ToolBar.h"
#include "MainWindow.h"
#include "TimeSyncItemEngineManager.h"
#include "LazyCaller.h"
#include <boost/scoped_ptr.hpp>
#include <set>
#include <stack>

using namespace std;
using namespace boost;
using namespace cnoid;

namespace cnoid {

    class ExtensionManagerImpl
    {
    public:
        ExtensionManagerImpl(const std::string& moduleName);
        ~ExtensionManagerImpl();

        void deleteManagedObjects();;
        
        scoped_ptr<MenuManager> menuManager;
        scoped_ptr<ItemManager> itemManager;
        scoped_ptr<TimeSyncItemEngineManager> timeSyncItemEngineManger;
        string moduleName;

        stack<ExtensionManager::PtrHolderBase*> pointerHolders;

        boost::signal<void()> sigSystemUpdated;
        boost::signal<void()> sigReleaseRequest;

    };

}

namespace {
    
    set<ExtensionManagerImpl*> extensionManagerImpls;

    void emitSigSystemUpdated()
    {
        set<ExtensionManagerImpl*>::iterator p = extensionManagerImpls.begin();
        while(p != extensionManagerImpls.end()){
            (*p)->sigSystemUpdated();
            ++p;
        }
    }

    LazyCaller emitSigSystemUpdatedCaller(emitSigSystemUpdated);
}


ExtensionManager::ExtensionManager(const std::string& moduleName)
{
    impl = new ExtensionManagerImpl(moduleName);
}


ExtensionManagerImpl::ExtensionManagerImpl(const std::string& moduleName)
    : moduleName(moduleName)
{
    extensionManagerImpls.insert(this);
}

ExtensionManager::~ExtensionManager()
{
    delete impl;
}


ExtensionManagerImpl::~ExtensionManagerImpl()
{
    ProjectManager::instance()->disconnectArchivers(moduleName);
    deleteManagedObjects();
    sigReleaseRequest();
    extensionManagerImpls.erase(this);
}


ItemManager& ExtensionManager::itemManager()
{
    if(!impl->itemManager){
        impl->itemManager.reset(new ItemManager(impl->moduleName, menuManager()));
    }
    return *impl->itemManager;
}


TimeSyncItemEngineManager& ExtensionManager::timeSyncItemEngineManger()
{
    if(!impl->timeSyncItemEngineManger){
        impl->timeSyncItemEngineManger.reset(new TimeSyncItemEngineManager(impl->moduleName));
    }
    return *impl->timeSyncItemEngineManger;
}


MenuManager& ExtensionManager::menuManager()
{
    if(!impl->menuManager){
        impl->menuManager.reset(new MenuManager(MainWindow::instance()->menuBar()));
    }
    return *impl->menuManager;
}


OptionManager& ExtensionManager::optionManager()
{
    static OptionManager optionManager;
    return optionManager;
}


ExtensionManager::PtrHolderBase::~PtrHolderBase()
{

}


void ExtensionManager::manageSub(PtrHolderBase* holder)
{
    impl->pointerHolders.push(holder);
}


void ExtensionManager::addView(View* view)
{
    manage(view);
    MainWindow::instance()->addView(impl->moduleName, view);
}


void ExtensionManager::addToolBar(ToolBar* toolBar)
{
    manage(toolBar);
    MainWindow::instance()->addToolBar(toolBar);
}


void ExtensionManagerImpl::deleteManagedObjects()
{
    while(!pointerHolders.empty()){
        ExtensionManager::PtrHolderBase* holder = pointerHolders.top();
        delete holder;
        pointerHolders.pop();
    }
}


SignalProxy< boost::signal<void()> > ExtensionManager::sigSystemUpdated()
{
    return impl->sigSystemUpdated;
}


void ExtensionManager::notifySystemUpdate()
{
    emitSigSystemUpdatedCaller.request();
}


SignalProxy< boost::signal<void()> > ExtensionManager::sigReleaseRequest()
{
    return impl->sigReleaseRequest;
}


void ExtensionManager::connectProjectArchiver(
    const std::string& name,
    boost::function<bool(Archive&)> storeFunction,
    boost::function<void(const Archive&)> restoreFunction)
{
    ProjectManager::instance()->connectArchiver(impl->moduleName, name, storeFunction, restoreFunction);
}
